@@ -31,17 +31,18 @@
 
 static GHashTable *window_hash = NULL;
 
-/* Keep 0-6 in sync with the numbers in the WindowState enum. Yeah I'm
+/* Keep 0-7 in sync with the numbers in the WindowState enum. Yeah I'm
  * a loser.
  */
-#define COMPRESS_STATE(window)                        \
-  ( ((window)->priv->is_minimized      << 0) |        \
-    ((window)->priv->is_maximized_horz << 1) |        \
-    ((window)->priv->is_maximized_vert << 2) |        \
-    ((window)->priv->is_shaded         << 3) |        \
-    ((window)->priv->skip_pager        << 4) |        \
-    ((window)->priv->skip_taskbar      << 5) |        \
-    ((window)->priv->is_sticky         << 6) )
+#define COMPRESS_STATE(window)                          \
+  ( ((window)->priv->is_minimized        << 0) |        \
+    ((window)->priv->is_maximized_horz   << 1) |        \
+    ((window)->priv->is_maximized_vert   << 2) |        \
+    ((window)->priv->is_shaded           << 3) |        \
+    ((window)->priv->skip_pager          << 4) |        \
+    ((window)->priv->skip_taskbar        << 5) |        \
+    ((window)->priv->is_sticky           << 6) |        \
+    ((window)->priv->is_hidden           << 7) )
 
 struct _WnckWindowPrivate
 {
@@ -70,7 +71,12 @@ struct _WnckWindowPrivate
   int y;
   int width;
   int height;
-
+  
+  /* true if transient_for points to root window,
+   * not another app window
+   */
+  guint transient_for_root : 1;
+  
   /* window state */
   guint is_minimized : 1;
   guint is_maximized_horz : 1;
@@ -79,12 +85,13 @@ struct _WnckWindowPrivate
   guint skip_pager : 1;
   guint skip_taskbar : 1;
   guint is_sticky : 1;
+  guint is_hidden : 1;
 
   /* _NET_WM_STATE_HIDDEN doesn't map directly into an
    * externally-visible state (it determines the WM_STATE
    * interpretation)
    */
-  guint net_wm_state_hidden : 1;
+  guint net_wm_state_hidden : 1;  
   guint wm_state_iconic : 1;
   
   /* idle handler for updates */
@@ -101,7 +108,7 @@ struct _WnckWindowPrivate
   guint need_emit_icon_changed : 1;
   guint need_update_actions : 1;
   guint need_update_wintype : 1;
-  guint need_update_transient_for : 1;
+  guint need_update_transient_for : 1;  
 };
 
 enum {
@@ -180,7 +187,7 @@ wnck_window_init (WnckWindow *window)
   window->priv = g_new0 (WnckWindowPrivate, 1);
 
   window->priv->name = g_strdup (FALLBACK_NAME);
-  window->priv->icon_name = g_strdup (FALLBACK_NAME);
+  window->priv->icon_name = NULL;
   window->priv->workspace = ALL_WORKSPACES;
 
   window->priv->icon_cache = _wnck_icon_cache_new ();
@@ -933,6 +940,85 @@ wnck_window_is_active (WnckWindow *window)
   return window == wnck_screen_get_active_window (window->priv->screen);
 }
 
+
+static WnckWindow*
+find_last_transient_for (GList *windows,
+                         Window xwindow)
+{
+  GList *tmp; 
+  WnckWindow *retval;
+
+  /* find _last_ transient for xwindow in the list */
+  
+  retval = NULL;
+  
+  tmp = windows;
+  while (tmp != NULL)
+    {
+      WnckWindow *w = tmp->data;
+
+      if (w->priv->transient_for == xwindow)
+        retval = w;
+      
+      tmp = tmp->next;
+    }
+
+  return retval;
+}
+
+/**
+ * wnck_window_activate_transient:
+ * @window: a #WnckWindow
+ *
+ * If @window has transients, activates the most likely transient
+ * instead of the window itself. Otherwise activates @window.
+ * 
+ * FIXME the ideal behavior of this function is probably to activate
+ * the most recently active window among @window and its transients.
+ * This is probably best implemented on the window manager side.
+ * 
+ **/
+void
+wnck_window_activate_transient (WnckWindow *window)
+{
+  GList *windows;
+  WnckWindow *transient;
+  WnckWindow *next;
+  
+  g_return_if_fail (WNCK_IS_WINDOW (window));
+
+  windows = wnck_screen_get_windows_stacked (window->priv->screen);
+
+  transient = NULL;
+  next = find_last_transient_for (windows, window->priv->xwindow);
+  
+  while (next != NULL)
+    {
+      if (next == window)
+        {
+          /* catch transient cycles */
+          transient = NULL;
+          break;
+        }
+
+      transient = next;
+      
+      next = find_last_transient_for (windows, transient);
+    }
+
+  if (transient != NULL)
+    {
+      /* Raise but don't focus the main window */
+      XRaiseWindow (gdk_display, window->priv->xwindow);
+      /* then activate the transient */
+      wnck_window_activate (transient);
+    }
+  else
+    {
+      wnck_window_activate (window);
+    }
+}
+
 static void
 get_icons (WnckWindow *window)
 {
@@ -1098,8 +1184,7 @@ wnck_window_is_visible_on_workspace (WnckWindow    *window,
   
   state = wnck_window_get_state (window);
 
-  if (state & (WNCK_WINDOW_STATE_MINIMIZED |
-               WNCK_WINDOW_STATE_SHADED))
+  if (state & WNCK_WINDOW_STATE_HIDDEN)
     return FALSE; /* not visible */
 
   return wnck_window_is_on_workspace (window, workspace);
@@ -1122,8 +1207,8 @@ wnck_window_set_icon_geometry (WnckWindow *window,
 			       int         width,
 			       int         height)
 {
-	_wnck_set_icon_geometry (window->priv->xwindow,
-				 x, y, width, height);
+  _wnck_set_icon_geometry (window->priv->xwindow,
+                           x, y, width, height);
 }
 
 /**
@@ -1338,13 +1423,21 @@ update_state (WnckWindow *window)
     {
     case WNCK_WINDOW_DESKTOP:
     case WNCK_WINDOW_DOCK:
+    case WNCK_WINDOW_SPLASHSCREEN:
+      window->priv->skip_taskbar = TRUE;
+      break;
+
     case WNCK_WINDOW_TOOLBAR:
     case WNCK_WINDOW_MENU:
     case WNCK_WINDOW_UTILITY:
-    case WNCK_WINDOW_SPLASHSCREEN:
     case WNCK_WINDOW_DIALOG:
     case WNCK_WINDOW_MODAL_DIALOG:
-      window->priv->skip_taskbar = TRUE;
+      /* Skip taskbar if the window is transient
+       * for some main application window
+       */
+      if (window->priv->transient_for != None &&
+          !window->priv->transient_for_root)
+        window->priv->skip_taskbar = TRUE;
       break;
       
     case WNCK_WINDOW_NORMAL:
@@ -1367,13 +1460,25 @@ update_state (WnckWindow *window)
     case WNCK_WINDOW_UTILITY:
       break;
     }
-
+  
   /* FIXME we need to recompute this if the window manager changes */
   if (wnck_screen_net_wm_supports (window->priv->screen,
                                    "_NET_WM_STATE_HIDDEN"))
-    window->priv->is_minimized = window->priv->net_wm_state_hidden;
+    {
+      window->priv->is_hidden = window->priv->net_wm_state_hidden;
+
+      /* FIXME this is really broken; need to bring it up on
+       * wm-spec-list. It results in showing an "Unminimize" menu
+       * item on task list, for shaded windows.
+       */
+      window->priv->is_minimized = window->priv->is_hidden;
+    }
   else
-    window->priv->is_minimized = window->priv->wm_state_iconic;
+    {
+      window->priv->is_minimized = window->priv->wm_state_iconic;
+      
+      window->priv->is_hidden = window->priv->is_minimized || window->priv->is_shaded;
+    }
 }
 
 static void
@@ -1417,8 +1522,7 @@ update_icon_name (WnckWindow *window)
       _wnck_get_text_property (window->priv->xwindow,
                                XA_WM_ICON_NAME);
 
-  if (window->priv->icon_name == NULL)
-    window->priv->icon_name = g_strdup (FALLBACK_NAME);
+  /* no fallback, get_icon_name falls back to regular window title */
 }
 
 static void
@@ -1622,10 +1726,16 @@ update_transient_for (WnckWindow *window)
                         &parent))
     {
       window->priv->transient_for = parent;
+
+      if (wnck_screen_get_for_root (window->priv->transient_for) != NULL)
+        window->priv->transient_for_root = TRUE;
+      else
+        window->priv->transient_for_root = FALSE;
     }
   else
     {
       window->priv->transient_for = None;
+      window->priv->transient_for_root = FALSE;
     }
 }
 
@@ -1667,7 +1777,8 @@ force_update_now (WnckWindow *window)
     window->priv->icon_name = old_icon_name;
   else
     {
-      if (strcmp (window->priv->icon_name, old_icon_name) != 0)
+      if (old_icon_name == NULL ||
+          strcmp (window->priv->icon_name, old_icon_name) != 0)
         emit_name_changed (window);
       g_free (old_icon_name);
     }
